package main.java.com.djrapitops.plan.api;

import java.util.UUID;
import main.java.com.djrapitops.plan.Plan;
import main.java.com.djrapitops.plan.data.AnalysisData;
import main.java.com.djrapitops.plan.data.UserData;
import main.java.com.djrapitops.plan.data.additional.AnalysisType;
import main.java.com.djrapitops.plan.data.additional.PluginData;
import main.java.com.djrapitops.plan.data.cache.DBCallableProcessor;
import main.java.com.djrapitops.plan.data.handling.info.HandlingInfo;
import main.java.com.djrapitops.plan.ui.DataRequestHandler;
import main.java.com.djrapitops.plan.ui.webserver.WebSocketServer;
import main.java.com.djrapitops.plan.utilities.HtmlUtils;
import main.java.com.djrapitops.plan.utilities.UUIDFetcher;
import static org.bukkit.Bukkit.getOfflinePlayer;
import org.bukkit.OfflinePlayer;

/**
 * This class contains the API methods.
 *
 * Methods can be called from Asyncronous task and are thread safe unless
 * otherwise stated.
 *
 * @author Rsl1122
 * @since 2.0.0
 * @see PluginData
 * @see AnalysisType
 * @see DBCallableProcessor
 * @see HandlingInfo
 */
public class API {

    private Plan plugin;

    /**
     * Class Construcor.
     *
     * @param plugin Current instance of Plan
     */
    public API(Plan plugin) {
        this.plugin = plugin;
    }

    /**
     * Check whether or not the plugin enabled successfully.
     *
     * @return true if plugin is enabled correctly.
     */
    public boolean isEnabled() {
        return plugin.isEnabled();
    }

    /**
     * Add a source of plugin data to the Plugins tab on Analysis and/or Inspect
     * page.
     *
     * Refer to documentation on github or Javadoc of PluginData to set-up a
     * data source that extends PluginData correctly.
     *
     * @param dataSource an object that extends PluginData-object, thus allowing
     * Analysis and Inspect to manage the data of a plugin correctly.
     * @see PluginData
     */
    public void addPluginDataSource(PluginData dataSource) {
        if (isEnabled()) {
            plugin.getHookHandler().addPluginDataSource(dataSource);
        }
    }

    /**
     * Used to get the link to InspectPage of a player.
     *
     * This method is useful if you have a table and want to link to the inspect
     * page.
     *
     * Html.LINK.parse("Link", "Playername") can be used to get a link
     * {@code <a href="Link">Playername</a>}
     *
     * @param name Playername of the player
     * @return ip:port/security/player/Playername
     */
    public String getPlayerInspectPageLink(String name) {
        return HtmlUtils.getInspectUrl(name);
    }

    /**
     * Schedule a UserData object to be fetched from the database or cache if
     * the player is online.
     *
     * The data will not be cached if it is not already cached.
     *
     * @param uuid UUID of the player.
     * @param processor Object implementing DBCallableProcessor, which
     * process(UserData data) method will be called.
     */
    public void scheduleForGet(UUID uuid, DBCallableProcessor processor) {
        plugin.getHandler().getUserDataForProcessing(processor, uuid, false);
    }

    /**
     * Schedule a HandlingInfo object to be processed.
     *
     * UserData associated with the UUID of the HandlingInfo object will be
     * cached.
     *
     * @param info object that extends HandlingInfo.
     */
    public void scheduleEventHandlingInfo(HandlingInfo info) {
        plugin.getHandler().addToPool(info);
    }

    /**
     * Used to cache a UserData object.
     *
     * If data is already cached it will be overridden.
     *
     * @param data UserData object. Will be placed to the data.getUuid() key in
     * the cache.
     */
    public void placeDataToCache(UserData data) {
        plugin.getHandler().cache(data);
    }

    /**
     * Used to save the cached data to the database.
     *
     * Should be only called from an Asyncronous thread.
     */
    public void saveCachedData() {
        plugin.getHandler().saveCachedUserData();
    }

    /**
     * Check if the UserData is cached to the InspectCache.
     *
     * @param uuid UUID of the player.
     * @return true/false
     */
    public boolean isPlayersDataInspectCached(UUID uuid) {
        return plugin.getInspectCache().isCached(uuid);
    }

    /**
     * Cache the UserData to InspectCache.
     *
     * Uses cache if data is cached or database if not. Call from an Asyncronous
     * thread.
     *
     * @param uuid UUID of the player.
     */
    public void cacheUserDataToInspectCache(UUID uuid) {
        plugin.getInspectCache().cache(uuid);
    }

    /**
     * Used to get the full Html of the Inspect page as a string.
     *
     * Check if the data is cached to InspectCache before calling this.
     *
     * @param uuid UUID of the player.
     * @return player.html with all placeholders replaced.
     */
    public String getPlayerHtmlAsString(UUID uuid) {
        WebSocketServer server = plugin.getUiServer();
        if (server != null) {
            return server.getDataReqHandler().getInspectHtml(uuid);
        }
        DataRequestHandler reqH = new DataRequestHandler(plugin);
        return reqH.getInspectHtml(uuid);
    }

    /**
     * Check if the Analysis has been run and is cached to the AnalysisCache.
     *
     * @return true/false
     */
    public boolean isAnalysisCached() {
        return plugin.getAnalysisCache().isCached();
    }

    /**
     * Run's the analysis with the current data in the cache and fetches rest from
     * the database.
     *
     * Starts a new Asyncronous task to run the analysis.
     */
    public void updateAnalysisCache() {
        plugin.getAnalysisCache().updateCache();
    }

    /**
     * Used to get the full Html of the Analysis page as a string.
     *
     * Check if the data is cached to AnalysisCache before calling this.
     *
     * @return analysis.html with all placeholders replaced.
     */
    public String getAnalysisHtmlAsString() {
        WebSocketServer server = plugin.getUiServer();
        if (server != null) {
            return server.getDataReqHandler().getAnalysisHtml();
        }
        DataRequestHandler reqH = new DataRequestHandler(plugin);
        return reqH.getAnalysisHtml();
    }

    /**
     * Used to get the AnalysisData object.
     *
     * Check if the data is cached to AnalysisCache before calling this.
     *
     * @return AnalysisData object.
     * @see AnalysisData
     */
    public AnalysisData getAnalysisDataFromCache() {
        return plugin.getAnalysisCache().getData();
    }

    /**
     * Used to get the playerName of a player who has played on the server.
     *
     * @param uuid UUID of the player.
     * @return Playername, eg "Rsl1122"
     * @throws IllegalStateException If the player has not played on the server
     * before.
     */
    public String getPlayerName(UUID uuid) throws IllegalStateException {
        OfflinePlayer offlinePlayer = getOfflinePlayer(uuid);
        if (offlinePlayer.hasPlayedBefore()) {
            return offlinePlayer.getName();
        }
        throw new IllegalStateException("Player has not played on this server before.");
    }

    /**
     * Uses UUIDFetcher to turn PlayerName to UUID.
     *
     * @param playerName Player's name
     * @return UUID of the Player
     * @throws Exception if player's name is not registered at Mojang
     */
    public UUID playerNameToUUID(String playerName) throws Exception {
        return UUIDFetcher.getUUIDOf(playerName);
    }
}
