package main.java.com.djrapitops.plan.database.databases;

import java.net.InetAddress;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;
import main.java.com.djrapitops.plan.Log;
import main.java.com.djrapitops.plan.Plan;
import main.java.com.djrapitops.plan.data.*;
import main.java.com.djrapitops.plan.data.cache.DBCallableProcessor;
import main.java.com.djrapitops.plan.database.Database;
import main.java.com.djrapitops.plan.database.tables.*;
import org.bukkit.GameMode;
import org.bukkit.Location;
import org.bukkit.scheduler.BukkitRunnable;
import static org.bukkit.Bukkit.getOfflinePlayer;

/**
 *
 * @author Rsl1122
 */
public abstract class SQLDB extends Database {

    final Plan plugin;

    private final boolean supportsModification;

    private Connection connection;

    /**
     *
     * @param plugin
     * @param supportsModification
     */
    public SQLDB(Plan plugin, boolean supportsModification) {
        super(plugin);
        this.plugin = plugin;
        this.supportsModification = supportsModification;
        boolean usingMySQL = getName().equals("MySQL");

        usersTable = new UsersTable(this, usingMySQL);
        gmTimesTable = new GMTimesTable(this, usingMySQL);
        sessionsTable = new SessionsTable(this, usingMySQL);
        killsTable = new KillsTable(this, usingMySQL);
        locationsTable = new LocationsTable(this, usingMySQL);
        ipsTable = new IPsTable(this, usingMySQL);
        nicknamesTable = new NicknamesTable(this, usingMySQL);
        commandUseTable = new CommandUseTable(this, usingMySQL);
        versionTable = new VersionTable(this, usingMySQL);

        startConnectionPingTask(plugin);
    }

    /**
     *
     * @param plugin
     * @throws IllegalArgumentException
     * @throws IllegalStateException
     */
    public void startConnectionPingTask(Plan plugin) throws IllegalArgumentException, IllegalStateException {
        // Maintains Connection.
        new BukkitRunnable() {
            @Override
            public void run() {
                try {
                    if (connection != null && !connection.isClosed()) {
                        connection.createStatement().execute("/* ping */ SELECT 1");
                    }
                } catch (SQLException e) {
                    connection = getNewConnection();
                }
            }
        }.runTaskTimerAsynchronously(plugin, 60 * 20, 60 * 20);
    }

    /**
     *
     * @return
     */
    @Override
    public boolean init() {
        super.init();
        try {
            return checkConnection();
        } catch (SQLException e) {
            Log.toLog(this.getClass().getName(), e);
            return false;
        }
    }

    /**
     *
     * @return @throws SQLException
     */
    public boolean checkConnection() throws SQLException {
        if (connection == null || connection.isClosed()) {
            connection = getNewConnection();

            if (connection == null || connection.isClosed()) {
                return false;
            }

            boolean newDatabase = true;
            try {
                getVersion();
                newDatabase = false;
            } catch (Exception e) {
            }
            if (!versionTable.createTable()) {
                Log.error("Failed to create table: " + versionTable.getTableName());
                return false;
            }
            if (newDatabase) {
                Log.info("New Database created.");
                setVersion(3);
            }
            for (Table table : getAllTables()) {
                if (!table.createTable()) {
                    Log.error("Failed to create table: " + table.getTableName());
                    return false;
                }
            }
            if (!newDatabase && getVersion() < 3) {
                setVersion(3);
            }
        }
        return true;
    }

    /**
     *
     * @return
     */
    public Table[] getAllTables() {
        return new Table[]{usersTable, locationsTable, gmTimesTable, ipsTable, nicknamesTable, sessionsTable, killsTable, commandUseTable};
    }

    /**
     *
     * @return
     */
    public Table[] getAllTablesInRemoveOrder() {
        return new Table[]{locationsTable, gmTimesTable, ipsTable, nicknamesTable, sessionsTable, killsTable, usersTable, commandUseTable};
    }

    /**
     *
     * @return
     */
    public abstract Connection getNewConnection();

    /**
     *
     * @throws SQLException
     */
    @Override
    public void close() throws SQLException {
        if (connection != null) {
            connection.close();
        }
    }

    /**
     *
     * @return @throws SQLException
     */
    @Override
    public int getVersion() throws SQLException {
        return versionTable.getVersion();
    }

    /**
     *
     * @param version
     * @throws SQLException
     */
    @Override
    public void setVersion(int version) throws SQLException {
        versionTable.setVersion(version);
    }

    /**
     *
     * @param uuid
     * @return
     */
    @Override
    public boolean wasSeenBefore(UUID uuid) {
        try {
            return usersTable.getUserId(uuid.toString()) != -1;
        } catch (SQLException e) {
            Log.toLog(this.getClass().getName(), e);
            return false;
        }
    }

    /**
     *
     * @param uuid
     * @return
     * @throws SQLException
     */
    @Override
    public boolean removeAccount(String uuid) throws SQLException {
        try {
            checkConnection();
        } catch (Exception e) {
            Log.toLog(this.getClass().getName(), e);
            return false;
        }
        int userId = usersTable.getUserId(uuid);
        if (userId == -1) {
            return false;
        }
        return locationsTable.removeUserLocations(userId)
                && ipsTable.removeUserIps(userId)
                && nicknamesTable.removeUserNicknames(userId)
                && gmTimesTable.removeUserGMTimes(userId)
                && sessionsTable.removeUserSessions(userId)
                && killsTable.removeUserKillsAndVictims(userId)
                && usersTable.removeUser(uuid);
    }

    /**
     *
     * @param uuid
     * @param processors
     * @throws SQLException
     */
    @Override
    public void giveUserDataToProcessors(UUID uuid, Collection<DBCallableProcessor> processors) throws SQLException {
        try {
            checkConnection();
        } catch (Exception e) {
            Log.toLog("Preparing for Exception report - Processors: " + processors.toString());
            Log.toLog(this.getClass().getName(), e);
            return;
        }
        // Check if user is in the database
        if (!wasSeenBefore(uuid)) {
            return;
        }
        // Get the data
        UserData data = new UserData(getOfflinePlayer(uuid), new DemographicsData());
        usersTable.addUserInformationToUserData(data);

        int userId = usersTable.getUserId(uuid);

        List<String> nicknames = nicknamesTable.getNicknames(userId);
        data.addNicknames(nicknames);
        if (nicknames.size() > 0) {
            data.setLastNick(nicknames.get(nicknames.size() - 1));
        }

        List<InetAddress> ips = ipsTable.getIPAddresses(userId);
        data.addIpAddresses(ips);

        HashMap<GameMode, Long> times = gmTimesTable.getGMTimes(userId);
        data.setGmTimes(times);
        List<SessionData> sessions = sessionsTable.getSessionData(userId);
        data.addSessions(sessions);
        data.setPlayerKills(killsTable.getPlayerKills(userId));
        for (DBCallableProcessor processor : processors) {
            processor.process(data);
        }
    }

    @Override
    public List<UserData> getUserDataForUUIDS(Collection<UUID> uuids) throws SQLException {
        if (uuids == null || uuids.isEmpty()) {
            return new ArrayList<>();
        }
        Map<UUID, Integer> userIds = usersTable.getAllUserIds();
        List<UserData> data = new ArrayList<>();
        for (UUID uuid : uuids) {
            if (!userIds.keySet().contains(uuid)) {
                continue;
            }
            UserData uData = new UserData(getOfflinePlayer(uuid), new DemographicsData());
            data.add(uData);
        }
        if (data.isEmpty()) {
            return data;
        }
        usersTable.addUserInformationToUserData(data);
        Map<Integer, UUID> idUuidRel = userIds.entrySet().stream().collect(Collectors.toMap(Map.Entry::getValue, Map.Entry::getKey));
        List<Integer> ids = userIds.entrySet().stream().filter(e -> uuids.contains(e.getKey())).map(e -> e.getValue()).collect(Collectors.toList());
        Log.debug("Ids: "+ids);
        Map<Integer, List<String>> nicknames = nicknamesTable.getNicknames(ids);
        Map<Integer, Set<InetAddress>> ipList = ipsTable.getIPList(ids);
        Map<Integer, List<KillData>> playerKills = killsTable.getPlayerKills(ids, idUuidRel);
        Map<Integer, List<SessionData>> sessionData = sessionsTable.getSessionData(ids);
        Log.debug("Sizes: U:"+uuids.size()+" D:"+data.size()+" I:"+userIds.size()+" N:"+nicknames.size()+" I:"+ipList.size()+" K:"+playerKills.size()+" S:"+sessionData.size());
        for (UserData uData : data) {
            UUID uuid = uData.getUuid();
            Integer id = userIds.get(uuid);
            uData.addIpAddresses(ipList.get(id));
            uData.addNicknames(nicknames.get(id));
            uData.addSessions(sessionData.get(id));
            uData.setPlayerKills(playerKills.get(id));
            uData.setGmTimes(gmTimesTable.getGMTimes(id));
        }
        return data;
    }

    /**
     *
     * @param data
     * @throws SQLException
     */
    @Override
    public void saveMultipleUserData(Collection<UserData> data) throws SQLException {
        checkConnection();
        if (data.isEmpty()) {
            return;
        }
        Set<Throwable> exceptions = new HashSet<>();
        List<UserData> saveLast = usersTable.saveUserDataInformationBatch(data);
        data.removeAll(saveLast);
        // Transform to map
        Map<UUID, UserData> userDatas = data.stream().collect(Collectors.toMap(UserData::getUuid, Function.identity()));
        // Get UserIDs
        Map<UUID, Integer> userIds = usersTable.getAllUserIds();
        // Empty dataset
        Map<Integer, List<Location>> locations = new HashMap<>();
        Map<Integer, Set<String>> nicknames = new HashMap<>();
        Map<Integer, String> lastNicks = new HashMap<>();
        Map<Integer, Set<InetAddress>> ips = new HashMap<>();
        Map<Integer, List<KillData>> kills = new HashMap<>();
        Map<Integer, UUID> uuids = userIds.entrySet().stream().collect(Collectors.toMap(Map.Entry::getValue, Map.Entry::getKey));
        Map<Integer, List<SessionData>> sessions = new HashMap<>();
        Map<Integer, Map<GameMode, Long>> gmTimes = new HashMap<>();
        // Put to dataset
        for (UUID uuid : userDatas.keySet()) {
            Integer id = userIds.get(uuid);
            UserData uData = userDatas.get(uuid);
            if (id == -1) {
                saveLast.add(uData);
                Log.debug("User not seen before, saving last: "+uuid);
                continue;
            }
            uData.access();
            locations.put(id, new ArrayList<>(uData.getLocations()));
            nicknames.put(id, new HashSet<>(uData.getNicknames()));
            lastNicks.put(id, uData.getLastNick());
            ips.put(id, new HashSet<>(uData.getIps()));
            kills.put(id, new ArrayList<>(uData.getPlayerKills()));
            sessions.put(id, new ArrayList<>(uData.getSessions()));
            gmTimes.put(id, uData.getGmTimes());
        }
        // Save
        locationsTable.saveAdditionalLocationsLists(locations);
        nicknamesTable.saveNickLists(nicknames, lastNicks);
        ipsTable.saveIPList(ips);
        killsTable.savePlayerKills(kills, uuids);
        sessionsTable.saveSessionData(sessions);
        for (Integer id : gmTimes.keySet()) {
            gmTimesTable.saveGMTimes(id, gmTimes.get(id));
        }
        for (Integer id : locations.keySet()) {
            UUID uuid = uuids.get(id);
            if (uuid != null) {
                UserData uData = userDatas.get(uuid);
                if (uData != null) {
                    uData.stopAccessing();
                }
            }
        }
        // Save leftovers
        for (UserData userData : saveLast) {
            try {
                saveUserData(userData);
            } catch (SQLException | NullPointerException e) {
                exceptions.add(e);
            }
        }
        if (!exceptions.isEmpty()) {
            Log.error("SEVERE: MULTIPLE ERRORS OCCURRED: " + exceptions.size());
            Log.toLog(this.getClass().getName(), exceptions);
        }
    }

    /**
     *
     * @param data
     * @throws SQLException
     */
    @Override
    public void saveUserData(UserData data) throws SQLException {
        if (data == null) {
            return;
        }
        UUID uuid = data.getUuid();
        if (uuid == null) {
            return;
        }
        checkConnection();
        Log.debug("DB_Save: " + data);
        data.access();
        usersTable.saveUserDataInformation(data);
        int userId = usersTable.getUserId(uuid.toString());
        sessionsTable.saveSessionData(userId, new ArrayList<>(data.getSessions()));
        locationsTable.saveAdditionalLocationsList(userId, data.getLocations());
        nicknamesTable.saveNickList(userId, new HashSet<>(data.getNicknames()), data.getLastNick());
        ipsTable.saveIPList(userId, new HashSet<>(data.getIps()));
        killsTable.savePlayerKills(userId, new ArrayList<>(data.getPlayerKills()));
        gmTimesTable.saveGMTimes(userId, data.getGmTimes());
        data.stopAccessing();
    }

    /**
     *
     * @param userId
     * @param locations
     * @throws SQLException
     */
    @Deprecated
    public void saveAdditionalLocationsList(int userId, List<Location> locations) throws SQLException {
        locationsTable.saveAdditionalLocationsList(userId, locations);
    }

    /**
     *
     * @param userId
     * @param names
     * @param lastNick
     * @throws SQLException
     */
    @Deprecated
    public void saveNickList(int userId, Set<String> names, String lastNick) throws SQLException {
        nicknamesTable.saveNickList(userId, names, lastNick);
    }

    /**
     *
     * @param userId
     * @param sessions
     * @throws SQLException
     * @deprecated Use sessionsTable instead.
     */
    @Deprecated
    public void saveSessionList(int userId, List<SessionData> sessions) throws SQLException {
        sessionsTable.saveSessionData(userId, sessions);
    }

    /**
     *
     * @param userId
     * @param kills
     * @throws SQLException
     */
    @Deprecated
    public void savePlayerKills(int userId, List<KillData> kills) throws SQLException {
        killsTable.savePlayerKills(userId, kills);
    }

    /**
     *
     * @param userId
     * @param ips
     * @throws SQLException
     */
    @Deprecated
    public void saveIPList(int userId, Set<InetAddress> ips) throws SQLException {
        ipsTable.saveIPList(userId, ips);
    }

    /**
     *
     * @param userId
     * @param gamemodeTimes
     * @throws SQLException
     */
    @Deprecated
    public void saveGMTimes(int userId, Map<GameMode, Long> gamemodeTimes) throws SQLException {
        gmTimesTable.saveGMTimes(userId, gamemodeTimes);
    }

    /**
     *
     */
    @Override
    public void clean() {
        try {
            checkConnection();
        } catch (SQLException e) {
            Log.toLog(this.getClass().getName(), e);
        }
    }

    /**
     *
     * @return
     */
    @Override
    public boolean removeAllData() {
        for (Table table : getAllTablesInRemoveOrder()) {
            if (!table.removeAllData()) {
                return false;
            }
        }
        return true;
    }

    /**
     *
     * @return
     */
    public boolean supportsModification() {
        return supportsModification;
    }

    /**
     *
     * @return
     */
    public Connection getConnection() {
        return connection;
    }
}
